@@ -28,13 +28,13 @@
 #include "gedit-debug.h"
 #include "gedit-document.h"
 #include "gedit-document-private.h"
-#include "gedit-multi-notebook.h"
 #include "gedit-notebook.h"
 #include "gedit-notebook-popup-menu.h"
 #include "gedit-tab.h"
 #include "gedit-tab-private.h"
 #include "gedit-utils.h"
 #include "gedit-commands-private.h"
+#include "gedit-window-private.h"
 
 typedef struct _GeditDocumentsGenericRow GeditDocumentsGenericRow;
 typedef struct _GeditDocumentsGenericRow GeditDocumentsGroupRow;
@@ -148,8 +148,6 @@ static const GtkTargetEntry panel_targets [] = {
 	{"GEDIT_DOCUMENTS_DOCUMENT_ROW", GTK_TARGET_SAME_APP, 0},
 };
 
-#define MAX_DOC_NAME_LENGTH 60
-
 #define ROW_OUTSIDE_LISTBOX -1
 
 static guint
@@ -439,22 +437,6 @@ group_row_refresh_visibility (GeditDocumentsPanel *panel)
 	gtk_widget_set_visible (first_group_row, !notebook_is_unique);
 }
 
-static gchar *
-doc_get_name (GeditDocument *doc)
-{
-	gchar *name;
-	gchar *docname;
-
-	name = tepl_file_get_short_name (tepl_buffer_get_file (TEPL_BUFFER (doc)));
-
-	/* Truncate the name so it doesn't get insanely wide. */
-	docname = tepl_utils_str_middle_truncate (name, MAX_DOC_NAME_LENGTH);
-
-	g_free (name);
-
-	return docname;
-}
-
 static void
 document_row_sync_tab_name_and_icon (GeditTab   *tab,
                                      GParamSpec *pspec,
@@ -467,7 +449,7 @@ document_row_sync_tab_name_and_icon (GeditTab   *tab,
 	GdkPixbuf *pixbuf;
 
 	doc = gedit_tab_get_document (tab);
-	name = doc_get_name (doc);
+	name = tepl_file_get_short_name (tepl_buffer_get_file (TEPL_BUFFER (doc)));
 
 	if (!gtk_text_buffer_get_modified (GTK_TEXT_BUFFER (doc)))
 	{
@@ -735,7 +717,7 @@ set_window (GeditDocumentsPanel *panel,
             GeditWindow         *window)
 {
 	panel->window = g_object_ref (window);
-	panel->mnb = GEDIT_MULTI_NOTEBOOK (_gedit_window_get_multi_notebook (window));
+	panel->mnb = _gedit_window_get_multi_notebook (window);
 
 	g_signal_connect (panel->mnb,
 	                  "notebook-removed",
@@ -928,7 +910,7 @@ panel_on_drag_begin (GtkWidget      *widget,
 
 	label = gtk_label_new (NULL);
 	gtk_label_set_markup (GTK_LABEL (label), name);
-	gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
+	gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_MIDDLE);
 	gtk_widget_set_halign (label, GTK_ALIGN_START);
 	gtk_widget_set_valign (label, GTK_ALIGN_CENTER);
 
@@ -1433,6 +1415,10 @@ gedit_documents_panel_init (GeditDocumentsPanel *panel)
 	panel->row_destination_index = ROW_OUTSIDE_LISTBOX;
 	panel->row_source_row_offset = 0;
 	panel->is_on_drag = FALSE;
+
+	/* Some widget config */
+	gtk_widget_set_hexpand (GTK_WIDGET (panel), TRUE);
+	gtk_widget_set_vexpand (GTK_WIDGET (panel), TRUE);
 }
 
 GtkWidget *
@@ -1558,7 +1544,7 @@ row_create (GtkWidget *row)
 	gtk_container_add (GTK_CONTAINER (event_box), generic_row->box);
 
 	generic_row->label = gtk_label_new (NULL);
-	gtk_label_set_ellipsize (GTK_LABEL (generic_row->label), PANGO_ELLIPSIZE_END);
+	gtk_label_set_ellipsize (GTK_LABEL (generic_row->label), PANGO_ELLIPSIZE_MIDDLE);
 	gtk_widget_set_halign (generic_row->label, GTK_ALIGN_START);
 	gtk_widget_set_valign (generic_row->label, GTK_ALIGN_CENTER);
 
