@@ -25,7 +25,6 @@
 
 #include <glib/gi18n.h>
 #include <libpeas/peas-extension-set.h>
-#include <tepl/tepl.h>
 
 #include "gedit-app.h"
 #include "gedit-app-private.h"
@@ -44,9 +43,8 @@
 #include "gedit-plugins-engine.h"
 #include "gedit-window-activatable.h"
 #include "gedit-enum-types.h"
-#include "gedit-status-menu-button.h"
 #include "gedit-settings.h"
-#include "gedit-menu-stack-switcher.h"
+#include "gedit-header-bar.h"
 
 struct _GeditWindowPrivate
 {
@@ -56,9 +54,7 @@ struct _GeditWindowPrivate
 
 	GeditMultiNotebook *multi_notebook;
 
-	GtkStack *side_panel;
-	GtkWidget *side_stack_switcher;
-	GtkWidget *side_panel_inline_stack_switcher;
+	GeditSidePanel *side_panel;
 	GtkWidget *bottom_panel;
 
 	GtkWidget *hpaned;
@@ -70,28 +66,22 @@ struct _GeditWindowPrivate
 	/* Widgets for fullscreen mode */
 	GtkWidget *fullscreen_eventbox;
 	GtkRevealer *fullscreen_revealer;
-	GtkWidget *fullscreen_headerbar;
-	GtkMenuButton *fullscreen_gear_button;
-	GtkMenuButton *fullscreen_open_recent_button;
+	GeditHeaderBar *fullscreen_headerbar;
 
 	/* statusbar and context ids for statusbar messages */
-	GtkWidget *statusbar;
+	GeditStatusbar *statusbar;
 	TeplOverwriteIndicator *overwrite_indicator;
 	TeplLineColumnIndicator *line_column_indicator;
-	GtkWidget *tab_width_button;
-	GtkWidget *language_button;
+	TeplStatusMenuButton *tab_width_button;
+	TeplStatusMenuButton *language_button;
 	GtkWidget *language_popover;
 	guint bracket_match_message_cid;
 	guint tab_width_id;
 	guint language_changed_id;
 
 	/* Headerbars */
-	GtkWidget *side_headerbar;
-	GtkWidget *headerbar;
-
-	GtkMenuButton *gear_button;
-
-	gint num_tabs_with_error;
+	GtkHeaderBar *side_headerbar;
+	GeditHeaderBar *headerbar;
 
 	gint width;
 	gint height;
@@ -177,11 +167,10 @@ gedit_window_get_property (GObject    *object,
 }
 
 static void
-save_panels_state (GeditWindow *window)
+save_side_panel_state (GeditWindow *window)
 {
-	const gchar *panel_page;
-
-	gedit_debug (DEBUG_WINDOW);
+	TeplPanelContainer *panel_container;
+	const gchar *item_name;
 
 	if (window->priv->side_panel_size > 0)
 	{
@@ -190,13 +179,20 @@ save_panels_state (GeditWindow *window)
 				    window->priv->side_panel_size);
 	}
 
-	panel_page = gtk_stack_get_visible_child_name (window->priv->side_panel);
-	if (panel_page != NULL)
+	panel_container = gedit_side_panel_get_panel_container (window->priv->side_panel);
+	item_name = tepl_panel_container_get_active_item_name (panel_container);
+	if (item_name != NULL)
 	{
 		g_settings_set_string (window->priv->window_settings,
 				       GEDIT_SETTINGS_SIDE_PANEL_ACTIVE_PAGE,
-				       panel_page);
+				       item_name);
 	}
+}
+
+static void
+save_bottom_panel_state (GeditWindow *window)
+{
+	const gchar *panel_page;
 
 	if (window->priv->bottom_panel_size > 0)
 	{
@@ -212,6 +208,13 @@ save_panels_state (GeditWindow *window)
 				       GEDIT_SETTINGS_BOTTOM_PANEL_ACTIVE_PAGE,
 				       panel_page);
 	}
+}
+
+static void
+save_panels_state (GeditWindow *window)
+{
+	save_side_panel_state (window);
+	save_bottom_panel_state (window);
 
 	g_settings_apply (window->priv->window_settings);
 }
@@ -284,8 +287,6 @@ gedit_window_dispose (GObject *object)
 	 */
 	peas_engine_garbage_collect (PEAS_ENGINE (gedit_plugins_engine_get_default ()));
 
-	g_clear_object (&window->priv->side_stack_switcher);
-
 	/* GTK+/GIO unref the action map in an idle. For the last GeditWindow,
 	 * the application quits before the idle, so the action map is not
 	 * unreffed, and some objects are not finalized on application shutdown
@@ -295,7 +296,9 @@ gedit_window_dispose (GObject *object)
 	 */
 	remove_actions (window);
 
-	window->priv->fullscreen_open_recent_button = NULL;
+	window->priv->side_headerbar = NULL;
+	window->priv->headerbar = NULL;
+	window->priv->fullscreen_headerbar = NULL;
 
 	G_OBJECT_CLASS (gedit_window_parent_class)->dispose (object);
 }
@@ -306,7 +309,7 @@ gedit_window_finalize (GObject *object)
 	GeditWindow *window = GEDIT_WINDOW (object);
 
 	g_free (window->priv->file_chooser_folder_uri);
-	g_slist_free_full (window->priv->closed_docs_stack, (GDestroyNotify)g_object_unref);
+	g_slist_free_full (window->priv->closed_docs_stack, g_object_unref);
 
 	G_OBJECT_CLASS (gedit_window_parent_class)->finalize (object);
 }
@@ -319,19 +322,7 @@ update_fullscreen (GeditWindow *window,
 
 	_gedit_multi_notebook_set_show_tabs (window->priv->multi_notebook, !is_fullscreen);
 
-	if (is_fullscreen)
-	{
-		gtk_widget_hide (window->priv->statusbar);
-	}
-	else
-	{
-		if (g_settings_get_boolean (window->priv->ui_settings, "statusbar-visible"))
-		{
-			gtk_widget_show (window->priv->statusbar);
-		}
-	}
-
-#ifndef OS_OSX
+#if !OS_MACOS
 	if (is_fullscreen)
 	{
 		gtk_widget_show_all (window->priv->fullscreen_eventbox);
@@ -523,12 +514,8 @@ gedit_window_class_init (GeditWindowClass *klass)
 	/* Bind class to template */
 	gtk_widget_class_set_template_from_resource (widget_class,
 	                                             "/org/gnome/gedit/ui/gedit-window.ui");
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, side_headerbar);
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, headerbar);
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, gear_button);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, hpaned);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, side_panel);
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, side_panel_inline_stack_switcher);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, vpaned);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, multi_notebook);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, bottom_panel);
@@ -537,8 +524,6 @@ gedit_window_class_init (GeditWindowClass *klass)
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, tab_width_button);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, fullscreen_eventbox);
 	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, fullscreen_revealer);
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, fullscreen_headerbar);
-	gtk_widget_class_bind_template_child_private (widget_class, GeditWindow, fullscreen_gear_button);
 }
 
 static void
@@ -871,15 +856,11 @@ setup_statusbar (GeditWindow *window)
 
 	gedit_debug (DEBUG_WINDOW);
 
+	_gedit_statusbar_set_window (window->priv->statusbar, window);
+
 	window->priv->bracket_match_message_cid = gtk_statusbar_get_context_id
 		(GTK_STATUSBAR (window->priv->statusbar), "bracket_match_message");
 
-	g_settings_bind (window->priv->ui_settings,
-	                 "statusbar-visible",
-	                 window->priv->statusbar,
-	                 "visible",
-	                 G_SETTINGS_BIND_GET);
-
 	/* Insert/Overwrite indicator */
 	window->priv->overwrite_indicator = tepl_overwrite_indicator_new ();
 	gtk_widget_show (GTK_WIDGET (window->priv->overwrite_indicator));
@@ -907,7 +888,8 @@ setup_statusbar (GeditWindow *window)
 	                                _gedit_app_get_tab_width_menu (GEDIT_APP (g_application_get_default ())));
 
 	/* Language button */
-	window->priv->language_popover = gtk_popover_new (window->priv->language_button);
+	gtk_widget_set_margin_end (GTK_WIDGET (window->priv->language_button), 3);
+	window->priv->language_popover = gtk_popover_new (GTK_WIDGET (window->priv->language_button));
 	gtk_menu_button_set_popover (GTK_MENU_BUTTON (window->priv->language_button),
 	                             window->priv->language_popover);
 
@@ -933,6 +915,8 @@ clone_window (GeditWindow *origin)
 	GeditWindow *window;
 	GdkScreen *screen;
 	GeditApp  *app;
+	TeplPanelContainer *origin_side_panel;
+	const gchar *side_panel_item_name;
 	const gchar *panel_page;
 
 	gedit_debug (DEBUG_WINDOW);
@@ -961,11 +945,14 @@ clone_window (GeditWindow *origin)
 	window->priv->side_panel_size = origin->priv->side_panel_size;
 	window->priv->bottom_panel_size = origin->priv->bottom_panel_size;
 
-	panel_page = gtk_stack_get_visible_child_name (origin->priv->side_panel);
-
-	if (panel_page)
+	origin_side_panel = gedit_side_panel_get_panel_container (origin->priv->side_panel);
+	side_panel_item_name = tepl_panel_container_get_active_item_name (origin_side_panel);
+	if (side_panel_item_name != NULL)
 	{
-		gtk_stack_set_visible_child_name (window->priv->side_panel, panel_page);
+		TeplPanelContainer *side_panel;
+
+		side_panel = gedit_side_panel_get_panel_container (window->priv->side_panel);
+		tepl_panel_container_set_active_item_name (side_panel, side_panel_item_name);
 	}
 
 	panel_page = gtk_stack_get_visible_child_name (GTK_STACK (origin->priv->bottom_panel));
@@ -986,36 +973,29 @@ clone_window (GeditWindow *origin)
 static void
 bracket_matched_cb (GtkSourceBuffer           *buffer,
 		    GtkTextIter               *iter,
-		    GtkSourceBracketMatchType  result,
+		    GtkSourceBracketMatchType  state,
 		    GeditWindow               *window)
 {
-	if (buffer != GTK_SOURCE_BUFFER (gedit_window_get_active_document (window)))
-		return;
+	gchar *msg;
 
-	switch (result)
+	if (buffer != GTK_SOURCE_BUFFER (gedit_window_get_active_document (window)))
 	{
-		case GTK_SOURCE_BRACKET_MATCH_NONE:
-			gtk_statusbar_pop (GTK_STATUSBAR (window->priv->statusbar),
-					   window->priv->bracket_match_message_cid);
-			break;
-		case GTK_SOURCE_BRACKET_MATCH_OUT_OF_RANGE:
-			gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
-						       window->priv->bracket_match_message_cid,
-						       _("Bracket match is out of range"));
-			break;
-		case GTK_SOURCE_BRACKET_MATCH_NOT_FOUND:
-			gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
-						       window->priv->bracket_match_message_cid,
-						       _("Bracket match not found"));
-			break;
-		case GTK_SOURCE_BRACKET_MATCH_FOUND:
-			gedit_statusbar_flash_message (GEDIT_STATUSBAR (window->priv->statusbar),
-						       window->priv->bracket_match_message_cid,
-						       _("Bracket match found on line: %d"),
-						       gtk_text_iter_get_line (iter) + 1);
-			break;
-		default:
-			g_assert_not_reached ();
+		return;
+	}
+
+	msg = gtk_source_utils_get_bracket_matched_message (iter, state);
+
+	if (msg != NULL)
+	{
+		gedit_statusbar_flash_message (window->priv->statusbar,
+					       window->priv->bracket_match_message_cid,
+					       msg);
+		g_free (msg);
+	}
+	else
+	{
+		gtk_statusbar_pop (GTK_STATUSBAR (window->priv->statusbar),
+				   window->priv->bracket_match_message_cid);
 	}
 }
 
@@ -1043,10 +1023,28 @@ overwrite_mode_changed (GtkTextView *view,
 	set_overwrite_mode (window, gtk_text_view_get_overwrite (view));
 }
 
+static void
+set_titles (GeditWindow *window,
+	    const gchar *single_title,
+	    const gchar *title,
+	    const gchar *subtitle)
+{
+	gedit_app_set_window_title (GEDIT_APP (g_application_get_default ()), window, single_title);
+
+	if (window->priv->headerbar != NULL)
+	{
+		gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->headerbar), title);
+		gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->headerbar), subtitle);
+	}
+
+	gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->fullscreen_headerbar), title);
+	gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->fullscreen_headerbar), subtitle);
+}
+
 #define MAX_TITLE_LENGTH 100
 
 static void
-set_title (GeditWindow *window)
+update_titles (GeditWindow *window)
 {
 	GeditTab *tab;
 	GeditDocument *doc = NULL;
@@ -1062,17 +1060,7 @@ set_title (GeditWindow *window)
 
 	if (tab == NULL)
 	{
-		gedit_app_set_window_title (GEDIT_APP (g_application_get_default ()),
-		                            window,
-		                            "gedit");
-		gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->headerbar),
-		                          "gedit");
-		gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->headerbar),
-		                             NULL);
-		gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->fullscreen_headerbar),
-		                          "gedit");
-		gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->fullscreen_headerbar),
-		                             NULL);
+		set_titles (window, "gedit", "gedit", NULL);
 		return;
 	}
 
@@ -1164,18 +1152,7 @@ set_title (GeditWindow *window)
 		}
 	}
 
-	gedit_app_set_window_title (GEDIT_APP (g_application_get_default ()),
-				    window,
-				    main_title);
-
-	gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->headerbar),
-	                          title);
-	gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->headerbar),
-	                             subtitle);
-	gtk_header_bar_set_title (GTK_HEADER_BAR (window->priv->fullscreen_headerbar),
-	                          title);
-	gtk_header_bar_set_subtitle (GTK_HEADER_BAR (window->priv->fullscreen_headerbar),
-	                             subtitle);
+	set_titles (window, main_title, title, subtitle);
 
 	g_free (dirname);
 	g_free (name);
@@ -1196,7 +1173,7 @@ tab_width_changed (GObject     *object,
 	new_tab_width = gtk_source_view_get_tab_width (GTK_SOURCE_VIEW (object));
 
 	label = g_strdup_printf (_("Tab Width: %u"), new_tab_width);
-	gedit_status_menu_button_set_label (GEDIT_STATUS_MENU_BUTTON (window->priv->tab_width_button), label);
+	tepl_status_menu_button_set_label_text (window->priv->tab_width_button, label);
 	g_free (label);
 }
 
@@ -1215,7 +1192,7 @@ language_changed (GObject     *object,
 	else
 		label = _("Plain Text");
 
-	gedit_status_menu_button_set_label (GEDIT_STATUS_MENU_BUTTON (window->priv->language_button), label);
+	tepl_status_menu_button_set_label_text (window->priv->language_button, label);
 
 	peas_extension_set_foreach (window->priv->extensions,
 	                            (PeasExtensionSetForeachFunc) extension_update_state,
@@ -1289,8 +1266,8 @@ update_statusbar (GeditWindow *window,
 						     TEPL_VIEW (new_view));
 		gtk_widget_show (GTK_WIDGET (window->priv->line_column_indicator));
 
-		gtk_widget_show (window->priv->tab_width_button);
-		gtk_widget_show (window->priv->language_button);
+		gtk_widget_show (GTK_WIDGET (window->priv->tab_width_button));
+		gtk_widget_show (GTK_WIDGET (window->priv->language_button));
 
 		window->priv->tab_width_id = g_signal_connect (new_view,
 							       "notify::tab-width",
@@ -1327,7 +1304,7 @@ tab_switched (GeditMultiNotebook *mnb,
 	if (new_tab == NULL || window->priv->dispose_has_run)
 		return;
 
-	set_title (window);
+	update_titles (window);
 	update_actions_sensitivity (window);
 
 	g_signal_emit (G_OBJECT (window),
@@ -1364,7 +1341,8 @@ analyze_tab_state (GeditTab    *tab,
 		case GEDIT_TAB_STATE_SAVING_ERROR:
 		case GEDIT_TAB_STATE_GENERIC_ERROR:
 			window->priv->state |= GEDIT_WINDOW_STATE_ERROR;
-			++window->priv->num_tabs_with_error;
+			break;
+
 		default:
 			/* NOP */
 			break;
@@ -1374,16 +1352,13 @@ analyze_tab_state (GeditTab    *tab,
 static void
 update_window_state (GeditWindow *window)
 {
-	GeditWindowState old_ws;
-	gint old_num_of_errors;
+	GeditWindowState old_window_state;
 
 	gedit_debug_message (DEBUG_WINDOW, "Old state: %x", window->priv->state);
 
-	old_ws = window->priv->state;
-	old_num_of_errors = window->priv->num_tabs_with_error;
+	old_window_state = window->priv->state;
 
 	window->priv->state = 0;
-	window->priv->num_tabs_with_error = 0;
 
 	gedit_multi_notebook_foreach_tab (window->priv->multi_notebook,
 					  (GtkCallback)analyze_tab_state,
@@ -1391,22 +1366,11 @@ update_window_state (GeditWindow *window)
 
 	gedit_debug_message (DEBUG_WINDOW, "New state: %x", window->priv->state);
 
-	if (old_ws != window->priv->state)
+	if (old_window_state != window->priv->state)
 	{
 		update_actions_sensitivity (window);
-
-		gedit_statusbar_set_window_state (GEDIT_STATUSBAR (window->priv->statusbar),
-						  window->priv->state,
-						  window->priv->num_tabs_with_error);
-
 		g_object_notify_by_pspec (G_OBJECT (window), properties[PROP_STATE]);
 	}
-	else if (old_num_of_errors != window->priv->num_tabs_with_error)
-	{
-		gedit_statusbar_set_window_state (GEDIT_STATUSBAR (window->priv->statusbar),
-						  window->priv->state,
-						  window->priv->num_tabs_with_error);
-	}
 }
 
 static void
@@ -1473,7 +1437,7 @@ sync_name (GeditTab    *tab,
 {
 	if (tab == gedit_window_get_active_tab (window))
 	{
-		set_title (window);
+		update_titles (window);
 		update_actions_sensitivity (window);
 	}
 }
@@ -1643,11 +1607,19 @@ drop_uris_cb (GtkWidget    *widget,
 static void
 update_fullscreen_revealer_state (GeditWindow *window)
 {
+	GtkMenuButton *open_recent_menu_button;
 	gboolean open_recent_menu_is_active;
-	gboolean hamburger_menu_is_active;
+	GtkMenuButton *hamburger_menu_button;
+	gboolean hamburger_menu_is_active = FALSE;
 
-	open_recent_menu_is_active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (window->priv->fullscreen_open_recent_button));
-	hamburger_menu_is_active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (window->priv->fullscreen_gear_button));
+	open_recent_menu_button = _gedit_header_bar_get_open_recent_menu_button (window->priv->fullscreen_headerbar);
+	open_recent_menu_is_active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (open_recent_menu_button));
+
+	hamburger_menu_button = _gedit_header_bar_get_hamburger_menu_button (window->priv->fullscreen_headerbar);
+	if (hamburger_menu_button != NULL)
+	{
+		hamburger_menu_is_active = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (hamburger_menu_button));
+	}
 
 	gtk_revealer_set_reveal_child (window->priv->fullscreen_revealer,
 				       (window->priv->in_fullscreen_eventbox ||
@@ -1973,13 +1945,13 @@ on_tab_removed (GeditMultiNotebook *multi,
 	g_return_if_fail (num_tabs >= 0);
 	if (num_tabs == 0)
 	{
-		set_title (window);
+		update_titles (window);
 
 		/* hide the additional widgets */
 		gtk_widget_hide (GTK_WIDGET (window->priv->overwrite_indicator));
 		gtk_widget_hide (GTK_WIDGET (window->priv->line_column_indicator));
-		gtk_widget_hide (window->priv->tab_width_button);
-		gtk_widget_hide (window->priv->language_button);
+		gtk_widget_hide (GTK_WIDGET (window->priv->tab_width_button));
+		gtk_widget_hide (GTK_WIDGET (window->priv->language_button));
 	}
 
 	if (!window->priv->dispose_has_run)
@@ -2171,11 +2143,11 @@ side_panel_visibility_changed (GtkWidget   *panel,
 				GEDIT_SETTINGS_SIDE_PANEL_VISIBLE,
 				visible);
 
-	/* sync the action state if the panel visibility was changed programmatically */
+	/* Sync the action state if the panel visibility was changed programmatically */
 	action = g_action_map_lookup_action (G_ACTION_MAP (window), "side-panel");
 	g_simple_action_set_state (G_SIMPLE_ACTION (action), g_variant_new_boolean (visible));
 
-	/* focus the right widget and set the right styles */
+	/* Focus the right widget */
 	if (visible)
 	{
 		gtk_widget_grab_focus (GTK_WIDGET (window->priv->side_panel));
@@ -2185,6 +2157,13 @@ side_panel_visibility_changed (GtkWidget   *panel,
 		gtk_widget_grab_focus (GTK_WIDGET (window->priv->multi_notebook));
 	}
 
+	/* Adapt the decoration layout if needed */
+	if (window->priv->headerbar == NULL ||
+	    window->priv->side_headerbar == NULL)
+	{
+		return;
+	}
+
 	g_object_get (gtk_settings_get_default (),
 		      "gtk-decoration-layout", &layout_desc,
 		      NULL);
@@ -2199,7 +2178,7 @@ side_panel_visibility_changed (GtkWidget   *panel,
 
 			layout_headerbar = g_strdup_printf ("%c%s", ':', tokens[1]);
 			gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (window->priv->headerbar), layout_headerbar);
-			gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (window->priv->side_headerbar), tokens[0]);
+			gtk_header_bar_set_decoration_layout (window->priv->side_headerbar, tokens[0]);
 
 			g_free (layout_headerbar);
 			g_strfreev (tokens);
@@ -2208,90 +2187,40 @@ side_panel_visibility_changed (GtkWidget   *panel,
 	else
 	{
 		gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (window->priv->headerbar), layout_desc);
-		gtk_header_bar_set_decoration_layout (GTK_HEADER_BAR (window->priv->side_headerbar), NULL);
+		gtk_header_bar_set_decoration_layout (window->priv->side_headerbar, NULL);
 	}
 
 	g_free (layout_desc);
 }
 
 static void
-on_side_panel_stack_children_number_changed (GtkStack    *stack,
-                                             GtkWidget   *widget,
-                                             GeditWindow *window)
+add_documents_panel (GeditWindow *window)
 {
-	GeditWindowPrivate *priv = window->priv;
-	GList *children;
+	GtkWidget *documents_panel;
+	TeplPanelContainer *panel_container;
+	TeplPanelItem *item;
 
-	children = gtk_container_get_children (GTK_CONTAINER (priv->side_panel));
+	documents_panel = gedit_documents_panel_new (window);
+	gtk_widget_show_all (documents_panel);
 
-	if (children != NULL && children->next != NULL)
-	{
-		gtk_widget_show (priv->side_stack_switcher);
-
-#if !INLINE_SIDE_PANEL_SWITCHER
-		gtk_header_bar_set_custom_title (GTK_HEADER_BAR (priv->side_headerbar), priv->side_stack_switcher);
-#endif
-	}
-	else
-	{
-		/* side_stack_switcher can get NULL in dispose, before stack children
-		   are being removed */
-		if (priv->side_stack_switcher != NULL)
-		{
-			gtk_widget_hide (priv->side_stack_switcher);
-		}
-
-#if !INLINE_SIDE_PANEL_SWITCHER
-		gtk_header_bar_set_custom_title (GTK_HEADER_BAR (priv->side_headerbar), NULL);
-#endif
-	}
-
-	g_list_free (children);
+	panel_container = gedit_side_panel_get_panel_container (window->priv->side_panel);
+	item = tepl_panel_add (TEPL_PANEL (panel_container),
+			       documents_panel,
+			       "GeditWindowDocumentsPanel",
+			       _("Documents"),
+			       NULL);
+	g_object_unref (item);
 }
 
 static void
 setup_side_panel (GeditWindow *window)
 {
-	GeditWindowPrivate *priv = window->priv;
-	GtkWidget *documents_panel;
+	g_signal_connect_after (window->priv->side_panel,
+				"notify::visible",
+				G_CALLBACK (side_panel_visibility_changed),
+				window);
 
-	gedit_debug (DEBUG_WINDOW);
-
-	g_signal_connect_after (priv->side_panel,
-	                        "notify::visible",
-	                        G_CALLBACK (side_panel_visibility_changed),
-	                        window);
-
-#if INLINE_SIDE_PANEL_SWITCHER
-	priv->side_stack_switcher = priv->side_panel_inline_stack_switcher;
-#else
-	priv->side_stack_switcher = gedit_menu_stack_switcher_new ();
-#endif
-
-	gtk_button_set_relief (GTK_BUTTON (priv->side_stack_switcher), GTK_RELIEF_NONE);
-	g_object_ref_sink (priv->side_stack_switcher);
-
-	gedit_utils_set_atk_name_description (priv->side_stack_switcher, _("Change side panel page"),  NULL);
-
-	gedit_menu_stack_switcher_set_stack (GEDIT_MENU_STACK_SWITCHER (priv->side_stack_switcher),
-	                                     priv->side_panel);
-
-	g_signal_connect (priv->side_panel,
-	                  "add",
-	                  G_CALLBACK (on_side_panel_stack_children_number_changed),
-	                  window);
-
-	g_signal_connect (priv->side_panel,
-	                  "remove",
-	                  G_CALLBACK (on_side_panel_stack_children_number_changed),
-	                  window);
-
-	documents_panel = gedit_documents_panel_new (window);
-	gtk_widget_show_all (documents_panel);
-	gtk_stack_add_titled (priv->side_panel,
-	                      documents_panel,
-	                      "GeditWindowDocumentsPanel",
-	                      _("Documents"));
+	add_documents_panel (window);
 }
 
 static void
@@ -2315,6 +2244,7 @@ bottom_panel_visibility_changed (GtkWidget   *panel_box,
 	/* focus the right widget */
 	if (visible)
 	{
+		/* FIXME: probably copy/paste error, should be the bottom panel. */
 		gtk_widget_grab_focus (GTK_WIDGET (window->priv->side_panel));
 	}
 	else
@@ -2376,21 +2306,16 @@ setup_bottom_panel (GeditWindow *window)
 static void
 init_side_panel_visibility (GeditWindow *window)
 {
-	gchar *child_name;
-	GtkWidget *child_widget;
+	TeplPanelContainer *side_panel;
+	gchar *item_name;
 	gboolean side_panel_visible;
 
-	gedit_debug (DEBUG_WINDOW);
+	side_panel = gedit_side_panel_get_panel_container (window->priv->side_panel);
 
-	child_name = g_settings_get_string (window->priv->window_settings,
-					    GEDIT_SETTINGS_SIDE_PANEL_ACTIVE_PAGE);
-	child_widget = gtk_stack_get_child_by_name (window->priv->side_panel, child_name);
-	if (child_widget != NULL)
-	{
-		gtk_stack_set_visible_child (window->priv->side_panel, child_widget);
-	}
-
-	g_free (child_name);
+	item_name = g_settings_get_string (window->priv->window_settings,
+					   GEDIT_SETTINGS_SIDE_PANEL_ACTIVE_PAGE);
+	tepl_panel_container_set_active_item_name (side_panel, item_name);
+	g_free (item_name);
 
 	side_panel_visible = g_settings_get_boolean (window->priv->ui_settings,
 						     GEDIT_SETTINGS_SIDE_PANEL_VISIBLE);
@@ -2546,14 +2471,27 @@ static void
 sync_fullscreen_actions (GeditWindow *window,
 			 gboolean     fullscreen)
 {
-	GtkMenuButton *button;
-	GPropertyAction *action;
+	GtkMenuButton *button = NULL;
+
+	if (fullscreen)
+	{
+		button = _gedit_header_bar_get_hamburger_menu_button (window->priv->fullscreen_headerbar);
+	}
+	else if (window->priv->headerbar != NULL)
+	{
+		button = _gedit_header_bar_get_hamburger_menu_button (window->priv->headerbar);
+	}
 
-	button = fullscreen ? window->priv->fullscreen_gear_button : window->priv->gear_button;
 	g_action_map_remove_action (G_ACTION_MAP (window), "hamburger-menu");
-	action = g_property_action_new ("hamburger-menu", button, "active");
-	g_action_map_add_action (G_ACTION_MAP (window), G_ACTION (action));
-	g_object_unref (action);
+
+	if (button != NULL)
+	{
+		GPropertyAction *action;
+
+		action = g_property_action_new ("hamburger-menu", button, "active");
+		g_action_map_add_action (G_ACTION_MAP (window), G_ACTION (action));
+		g_object_unref (action);
+	}
 }
 
 static void
@@ -2565,99 +2503,102 @@ init_amtk_application_window (GeditWindow *gedit_window)
 	amtk_application_window_set_statusbar (amtk_window, GTK_STATUSBAR (gedit_window->priv->statusbar));
 }
 
+#if GEDIT_HAS_HEADERBAR
 static void
-open_recent_menu_item_activated_cb (GtkRecentChooser *recent_chooser,
-				    gpointer          user_data)
+create_side_headerbar (GeditWindow *window)
 {
-	GeditWindow *window = GEDIT_WINDOW (user_data);
-	gchar *uri;
-	GFile *location;
+	TeplPanelContainer *panel_container;
+	TeplPanelSwitcherMenu *switcher;
+	GtkSizeGroup *size_group;
 
-	uri = gtk_recent_chooser_get_current_uri (recent_chooser);
-	location = g_file_new_for_uri (uri);
+	g_return_if_fail (window->priv->side_headerbar == NULL);
 
-	gedit_commands_load_location (window, location, NULL, 0, 0);
+	window->priv->side_headerbar = GTK_HEADER_BAR (gtk_header_bar_new ());
+	gtk_header_bar_set_show_close_button (window->priv->side_headerbar, TRUE);
 
-	g_free (uri);
-	g_object_unref (location);
-}
+	panel_container = gedit_side_panel_get_panel_container (window->priv->side_panel);
+	switcher = tepl_panel_switcher_menu_new (panel_container);
+	gtk_widget_show (GTK_WIDGET (switcher));
 
-static GtkWidget *
-create_open_buttons (GeditWindow    *window,
-		     GtkMenuButton **open_recent_button)
-{
-	GtkWidget *hbox;
-	GtkStyleContext *style_context;
-	GtkWidget *open_dialog_button;
-	GtkWidget *my_open_recent_button;
-	AmtkApplicationWindow *amtk_window;
-	GtkRecentChooserMenu *recent_menu;
+	gtk_header_bar_set_custom_title (window->priv->side_headerbar, GTK_WIDGET (switcher));
 
-	/* It currently needs to be a GtkBox, not a GtkGrid, because GtkGrid and
-	 * GTK_STYLE_CLASS_LINKED doesn't work as expected in a RTL locale.
-	 * Probably a GtkGrid bug.
+	g_object_bind_property (window->priv->side_panel, "visible",
+				window->priv->side_headerbar, "visible",
+				G_BINDING_SYNC_CREATE);
+
+	/* There are two horizontal GtkPaned, but one should not be able to have
+	 * a lower position than the other.
 	 */
-	hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 0);
-	style_context = gtk_widget_get_style_context (hbox);
-	gtk_style_context_add_class (style_context, GTK_STYLE_CLASS_LINKED);
-
-	open_dialog_button = gtk_button_new_with_mnemonic (_("_Open"));
-	gtk_widget_set_tooltip_text (open_dialog_button, _("Open a file"));
-	gtk_actionable_set_action_name (GTK_ACTIONABLE (open_dialog_button), "win.open");
-
-	my_open_recent_button = gtk_menu_button_new ();
-	gtk_widget_set_tooltip_text (my_open_recent_button, _("Open a recently used file"));
-
-	recent_menu = amtk_application_window_create_open_recent_menu_base ();
-
-	amtk_window = amtk_application_window_get_from_gtk_application_window (GTK_APPLICATION_WINDOW (window));
-	amtk_application_window_connect_recent_chooser_menu_to_statusbar (amtk_window, recent_menu);
-
-	g_signal_connect_object (recent_menu,
-				 "item-activated",
-				 G_CALLBACK (open_recent_menu_item_activated_cb),
-				 window,
-				 0);
-
-	gtk_menu_button_set_popup (GTK_MENU_BUTTON (my_open_recent_button),
-				   GTK_WIDGET (recent_menu));
-
-	gtk_container_add (GTK_CONTAINER (hbox), open_dialog_button);
-	gtk_container_add (GTK_CONTAINER (hbox), my_open_recent_button);
-	gtk_widget_show_all (hbox);
-
-	if (open_recent_button != NULL)
-	{
-		*open_recent_button = GTK_MENU_BUTTON (my_open_recent_button);
-	}
-
-	return hbox;
+	size_group = gtk_size_group_new (GTK_SIZE_GROUP_HORIZONTAL);
+	gtk_size_group_add_widget (size_group, GTK_WIDGET (window->priv->side_headerbar));
+	gtk_size_group_add_widget (size_group, GTK_WIDGET (window->priv->side_panel));
+	g_object_unref (size_group);
 }
 
 static void
-init_open_buttons (GeditWindow *window)
+create_titlebar (GeditWindow *window)
 {
-	gtk_container_add_with_properties (GTK_CONTAINER (window->priv->headerbar),
-					   create_open_buttons (window, NULL),
-					   "position", 0, /* The first on the left. */
-					   NULL);
+	GtkPaned *titlebar_hpaned;
 
-	gtk_container_add_with_properties (GTK_CONTAINER (window->priv->fullscreen_headerbar),
-					   create_open_buttons (window, &(window->priv->fullscreen_open_recent_button)),
-					   "position", 0, /* The first on the left. */
-					   NULL);
+	g_return_if_fail (window->priv->headerbar == NULL);
 
-	g_signal_connect (GTK_TOGGLE_BUTTON (window->priv->fullscreen_open_recent_button),
-	                  "toggled",
-	                  G_CALLBACK (on_fullscreen_toggle_button_toggled),
-	                  window);
+	create_side_headerbar (window);
+
+	window->priv->headerbar = _gedit_header_bar_new (window, FALSE);
+	gtk_widget_show (GTK_WIDGET (window->priv->headerbar));
+	gtk_header_bar_set_show_close_button (GTK_HEADER_BAR (window->priv->headerbar), TRUE);
+
+	titlebar_hpaned = GTK_PANED (gtk_paned_new (GTK_ORIENTATION_HORIZONTAL));
+	gtk_widget_show (GTK_WIDGET (titlebar_hpaned));
+
+	gtk_paned_pack1 (titlebar_hpaned,
+			 GTK_WIDGET (window->priv->side_headerbar),
+			 FALSE, FALSE);
+	gtk_paned_pack2 (titlebar_hpaned,
+			 GTK_WIDGET (window->priv->headerbar),
+			 TRUE, FALSE);
+
+	g_object_bind_property (window->priv->hpaned, "position",
+				titlebar_hpaned, "position",
+				G_BINDING_BIDIRECTIONAL | G_BINDING_SYNC_CREATE);
+
+	gtk_window_set_titlebar (GTK_WINDOW (window), GTK_WIDGET (titlebar_hpaned));
+}
+#endif /* GEDIT_HAS_HEADERBAR */
+
+static void
+create_fullscreen_headerbar (GeditWindow *window)
+{
+	GtkMenuButton *button;
+
+	g_return_if_fail (window->priv->fullscreen_headerbar == NULL);
+
+	window->priv->fullscreen_headerbar = _gedit_header_bar_new (window, TRUE);
+	gtk_widget_show (GTK_WIDGET (window->priv->fullscreen_headerbar));
+
+	gtk_container_add (GTK_CONTAINER (window->priv->fullscreen_revealer),
+			   GTK_WIDGET (window->priv->fullscreen_headerbar));
+
+	button = _gedit_header_bar_get_open_recent_menu_button (window->priv->fullscreen_headerbar);
+	g_signal_connect (button,
+			  "toggled",
+			  G_CALLBACK (on_fullscreen_toggle_button_toggled),
+			  window);
+
+	button = _gedit_header_bar_get_hamburger_menu_button (window->priv->fullscreen_headerbar);
+	if (button != NULL)
+	{
+		g_signal_connect (button,
+				  "toggled",
+				  G_CALLBACK (on_fullscreen_toggle_button_toggled),
+				  window);
+	}
 }
 
 static void
 gedit_window_init (GeditWindow *window)
 {
 	GtkTargetList *tl;
-	GMenuModel *hamburger_menu;
 
 	gedit_debug (DEBUG_WINDOW);
 
@@ -2680,8 +2621,14 @@ gedit_window_init (GeditWindow *window)
 	window->priv->message_bus = gedit_message_bus_new ();
 
 	gtk_widget_init_template (GTK_WIDGET (window));
+
 	init_amtk_application_window (window);
-	init_open_buttons (window);
+
+#if GEDIT_HAS_HEADERBAR
+	create_titlebar (window);
+#endif
+
+	create_fullscreen_headerbar (window);
 
 	g_action_map_add_action_entries (G_ACTION_MAP (window),
 	                                 win_entries,
@@ -2694,25 +2641,6 @@ gedit_window_init (GeditWindow *window)
 	setup_fullscreen_eventbox (window);
 	sync_fullscreen_actions (window, FALSE);
 
-	hamburger_menu = _gedit_app_get_hamburger_menu (GEDIT_APP (g_application_get_default ()));
-	if (hamburger_menu)
-	{
-		gtk_menu_button_set_menu_model (window->priv->gear_button, hamburger_menu);
-		gtk_menu_button_set_menu_model (window->priv->fullscreen_gear_button, hamburger_menu);
-	}
-	else
-	{
-		gtk_widget_hide (GTK_WIDGET (window->priv->gear_button));
-		gtk_widget_hide (GTK_WIDGET (window->priv->fullscreen_gear_button));
-		gtk_widget_set_no_show_all (GTK_WIDGET (window->priv->gear_button), TRUE);
-		gtk_widget_set_no_show_all (GTK_WIDGET (window->priv->fullscreen_gear_button), TRUE);
-	}
-
-	g_signal_connect (GTK_TOGGLE_BUTTON (window->priv->fullscreen_gear_button),
-	                  "toggled",
-	                  G_CALLBACK (on_fullscreen_toggle_button_toggled),
-	                  window);
-
 	/* Setup status bar */
 	setup_statusbar (window);
 
@@ -3185,20 +3113,27 @@ _gedit_window_is_removing_tabs (GeditWindow *window)
 	return window->priv->removing_tabs;
 }
 
+GeditSidePanel *
+_gedit_window_get_whole_side_panel (GeditWindow *window)
+{
+	g_return_val_if_fail (GEDIT_IS_WINDOW (window), NULL);
+
+	return window->priv->side_panel;
+}
+
 /**
  * gedit_window_get_side_panel:
  * @window: a #GeditWindow
  *
- * Gets the side panel of the @window.
- *
- * Returns: (transfer none): the side panel's #GtkStack.
+ * Returns: (transfer none): the side panel.
+ * Since: 46
  */
-GtkWidget *
+TeplPanel *
 gedit_window_get_side_panel (GeditWindow *window)
 {
 	g_return_val_if_fail (GEDIT_IS_WINDOW (window), NULL);
 
-	return GTK_WIDGET (window->priv->side_panel);
+	return TEPL_PANEL (gedit_side_panel_get_panel_container (window->priv->side_panel));
 }
 
 /**
@@ -3230,7 +3165,7 @@ gedit_window_get_statusbar (GeditWindow *window)
 {
 	g_return_val_if_fail (GEDIT_IS_WINDOW (window), NULL);
 
-	return window->priv->statusbar;
+	return GTK_WIDGET (window->priv->statusbar);
 }
 
 /**
